(* Alice versions
  import signature ORDERED from "x-alice:/lib/data/ORDERED-sig";
  import signature MAP from "x-alice:/lib/data/MAP-sig"
  import functor MkRedBlackMap from "x-alice:/lib/data/MkRedBlackMap";

  structure IntPair : ORDERED =
  struct
      type t = int * int
      fun compare ((a,b), (c,d)) =
            case Int.compare(a,c) of
       	    EQUAL => Int.compare(b,d)
            | other => other
  end;

  structure Binarymap =
    MkRedBlackMap (IntPair: ORDERED) :> MAP where type key = IntPair.t;

  val inputLine = Option.valOf o TextIO.inputLine;
  val sort = List.sort;
  val isSorted = List.isSorted;
  val mkDict = Binarymap.empty;
  fun insert ((r,c,v), bmap) = Binarymap.insert (bmap, (r,c), v);
  fun peek (ps, (r,c)) = Binarymap.lookup (ps, (r,c));
 *)

(* SML versions *)
  app load ["Binarymap","Int", "Listsort", "Time", "Timer"];

  val inputLine = TextIO.inputLine;
  val sort = Listsort.sort;
  val isSorted = Listsort.sorted;
  val mkDict = let 
                   fun compare ((a,b), (c,d)) =
                         case Int.compare(a,c) of
                  	   EQUAL => Int.compare(b,d)
                         | other => other
               in
                  Binarymap.mkDict compare
               end;
  fun insert ((r,c,v), bmap) = Binarymap.insert (bmap, (r,c), v);
  fun peek (pk, (r,c)) = Binarymap.peek (pk, (r,c));

(* end of versions *)

type row       = int
type col       = int
type value     = int
type nummines = int
type size      = row * col
type point     = row * col
type refpoint  = row * col * value
type board     = size * nummines * refpoint list
type minefield = point list;

exception Mines of string;

(* validTask board = true, if the board board is valid
   validTask : board -> bool
*)
fun validTask (size, noOfMines, refPL) =
    let
        val (maxR, maxC) = size
        val area = maxR * maxC
        val noOfRefPoints = length refPL
        fun validPoint (row, col) =
              1 <= row andalso 1 <= col andalso
              row <= maxR andalso col <= maxC
        fun validRefPoint (row, col, value)  =
              validPoint (row, col) andalso
              0 <= value andalso value <= 8;
        fun compareLexicographically ((x1,y1,_), (x2,y2,_)) =
              if      x1 < x2 orelse x1 = x2 andalso y1 < y2 then LESS
              else if x1 > x2 orelse x1 = x2 andalso y1 > y2 then GREATER
              else EQUAL
        fun differentNeighbours (rp1::(rps as rp2::_)) =
              compareLexicographically (rp1, rp2) <> EQUAL
              andalso differentNeighbours rps
          | differentNeighbours _ = true
    in 
        (* size of board and number of mines are valid: *)
        1 <= maxR andalso 1 <= maxC andalso
        area - noOfRefPoints >= noOfMines andalso
        (* all elements of refPL are valid: *)
        (List.all (fn x => x) (map validRefPoint refPL)) andalso
        (* refPL is lexicographically sorted: *)
        isSorted compareLexicographically refPL andalso
        (* each element of refPL refers to different coordinates; only
           neighbour elements should be compared for refPL is sorted: *)
        differentNeighbours refPL
    end;

(* minesIn fi = the board produced from the content of the text file fi
   minesIn : string -> board
*)
fun minesIn fi =
    let
	fun toPair [a,b] = (a,b)
	  | toPair _ =
	    raise Mines("Syntactically incorrect size description in file " ^ fi)

	fun toTriple [a,b,c] = (a,b,c)
	  | toTriple _ =
	    raise Mines("Syntactically incorrect field description in file " ^ fi)

	fun fmString l2n = l2n o
	                   (List.mapPartial Int.fromString) o
			   (String.tokens Char.isSpace)

	fun refPointsIn s =
	    if TextIO.endOfStream s then []
	    else (fmString toTriple) (inputLine s) :: refPointsIn s

	fun compare ((y1,x1,_), (y2,x2,_)) =
	    if      y1 > y2 then GREATER
	    else if y1 < y2 then LESS
	    else if x1 > x2 then GREATER
	    else if x1 < x2 then LESS
	    else raise Mines("Repeated coordinates in file " ^ fi)

	val istr      = TextIO.openIn fi
	val size     = (fmString toPair) (inputLine istr)
	val noOfMines  = Option.valOf(Int.fromString(inputLine istr))
	val refPoints = sort compare (refPointsIn istr)
	val board     = (size, noOfMines, refPoints) before TextIO.closeIn istr
    in
	if validTask board then board
	else raise Mines("Invalid board in file " ^ fi)
    end;

(* drawBoard (board, minefield) = draw the arrangement as described by
                                  the parameters board and minefield
   drawBoard : board * minefield -> string
*)
fun drawBoard (((size,amount), _, rps), mines) =
    let
	val p0s = mkDict
	val p1s = foldl insert p0s (map (fn (r,c) => (r,c,#"*")) mines)
	val p2s = foldl insert p1s (map (fn (r,c,v) => (r,c,chr(ord #"0" + v))) rps)

	val tab  = String.concat o List.tabulate
	val line = tab (amount, fn _ => "+---") ^ "+\n"

	fun drawCell r c = case peek(p2s, (r+1,c+1)) of
	                        SOME c => " " ^ str c ^ " |"
			      | NONE   => "   |"

	fun drawLine i = "|" ^ tab (amount, drawCell i) ^ "\n" ^ line
    in
	line ^ tab (size, drawLine)
    end

(* drawBoards (board, fields) = draws the boards, separated by empty lines, as
                                described by the parameters fields and board
   drawBoards : board * minefield list -> string
*)
fun drawBoards (board, fields) =
    foldl (fn (mines,s) => s ^ drawBoard(board, mines)) "" fields

(* minesOut(fo, b, fields) = converts the solutions list fields of the board
                             b to human readable form, and then writes this
                             into the text file fo
   minesOut : string * board * minefield list -> unit
*)
fun minesOut(fo, b, fields) =
      let val ostr = TextIO.openOut fo
	  val _ = TextIO.output(ostr, drawBoards(b, fields))
      in
	 TextIO.closeOut ostr
      end
  
(* solve(fi, fo) = reads a task from the text file fi, and writes its
                   solutions into the text file fo, while it makes use
                   of the function mines to solve the task
   solve : string * string -> unit
*)
fun solve (fi, fo) =
    let val board = minesIn fi
	val fields = mines board
	val _ = minesOut(fo, board, fields)
    in
	"Task: " ^ fi ^ ", " ^
	Int.toString(length fields) ^ " solution(s)\n"
    end
  
(* timer(fi, fo) = reads a task from the text file fi, and writes its
                   solutions into the text file fo, while it makes use
                   of the function mines to solve the task; the function
                   returns a string composed of the name fi, the number of
                   solutions, and the time needed to solve the task in seconds.
   timer : string * string -> string
*)

fun timer (fi, fo) =
    let val board = minesIn fi
	val start = Timer.startCPUTimer()
	val fields = mines board
	val {usr=tim,...} = Timer.checkCPUTimer start
	val _ = minesOut(fo, board, fields)
    in
	"Task: " ^ fi ^ ", " ^
	Int.toString(length fields) ^ " solution, " ^
	"time: " ^ Time.fmt 2 tim ^ " s\n"
    end
